from marshmallow import Schema, fields, post_load, EXCLUDE, INCLUDE
import logging

log = logging.getLogger('prom2teams')


class MessageSchema(Schema):
    class Meta:
        unknown = EXCLUDE
    def __init__(self, exclude_fields=tuple(), exclude_annotations=tuple()):
        super().__init__()
        self.exclude_fields = exclude_fields
        self.exclude_annotations = exclude_annotations

    receiver = fields.Str()
    status = fields.Str(default='unknown', missing='unknown')
    alerts = fields.Nested('AlertSchema', many=True)
    groupKey = fields.Str()
    externalURL = fields.Str()
    version = fields.Str()
    commonAnnotations = fields.Nested('AnnotationSchema', many=False, unknown=EXCLUDE)
    commonLabels = fields.Nested('LabelSchema', many=False, unknown=EXCLUDE)    
    groupLabels = fields.Nested('LabelSchema', many=False, unknown=EXCLUDE)    

    @post_load()
    def get_alerts(self, message):
        log.debug('JSON received is:\n%s', str(message))
        prom_alerts = []

        base_labels = ('alertname', 'device', 'fstype', 'instance', 'mountpoint', 'severity')
        excluded = base_labels + self.exclude_fields
        base_annotations = ('description', 'summary')
        excluded_annotations = base_annotations + self.exclude_annotations

        for alert in message['alerts']:
            status = alert['status']
            summary = alert['annotations']['summary']
            instance = alert['labels']['instance']
            name = alert['labels']['alertname']
            description = alert['annotations']['description']
            severity = alert['labels']['severity']
            extra_labels = dict()
            extra_annotations = dict()

            for key in alert['labels']:
                if key not in excluded:
                    extra_labels[key] = alert['labels'][key]

            for key in alert.get('annotations'):
                annotation = alert['annotations'][key]

                # Annotations with 1 or more dots in them will be interpreted by JSON methods beyond this file to be objects with subobjects.
                # Here, we consider only the dotless case - where an annotation is not interpreted as a dictionary.
                annotation_is_not_dict = not(isinstance(annotation, dict))

                if key not in excluded_annotations and annotation_is_not_dict:
                    extra_annotations[key] = annotation

            alert = PrometheusAlert(name, status, severity, summary, instance, description, extra_labels, extra_annotations)
            prom_alerts.append(alert)
        return prom_alerts


class AlertSchema(Schema):
    status = fields.Str(default='unknown', missing='unknown')
    labels = fields.Nested('LabelSchema', many=False, unknown=INCLUDE)
    annotations = fields.Nested('AnnotationSchema', many=False, unknown=INCLUDE)
    startsAt = fields.DateTime()
    endsAt = fields.DateTime()
    generatorURL = fields.Str()
    fingerprint = fields.Str()
    class Meta:
        unknown = EXCLUDE

class LabelSchema(Schema):
    alertname = fields.Str(default='unknown', missing='unknown')
    device = fields.Str()
    fstype = fields.Str()
    instance = fields.Str(default='unknown', missing='unknown')
    job = fields.Str()
    mountpoint = fields.Str()
    severity = fields.Str(default='unknown', missing='unknown')


class AnnotationSchema(Schema):
    description = fields.Str(default='unknown', missing='unknown')
    summary = fields.Str(default='unknown', missing='unknown')


class PrometheusAlert:
    def __init__(self, name, status, severity, summary, instance, description, extra_labels=None, extra_annotations=None):
        self.name = name
        self.status = status
        self.severity = severity
        self.summary = summary
        self.instance = instance
        self.description = description
        self.extra_labels = extra_labels
        self.extra_annotations = extra_annotations



import logging
from flask_restplus import Api

log = logging.getLogger(__name__)

api_v2 = Api(version='2.0', title='Prom2Teams API v2',
             description='A swagger interface for Prom2Teams webservices')


@api_v2.errorhandler
def default_error_handler(e):
    msg = 'An unhandled exception occurred.'
    log.exception(msg + e)
    return {'message': msg}, 500



from flask_restplus import fields
from . import api_v1

annotations = api_v1.model('annotations', {
    'description': fields.String(default='disk usage 93% on rootfs device'),
    'summary': fields.String(default='Disk usage alert on CS30.evilcorp')
})

labels = api_v1.model('labels', {
    'alertname': fields.String(default='DiskSpace'),
    'fstype': fields.String(default='rootfs'),
    'device': fields.String(default='rootfs'),
    'instance': fields.String(default='cs30.evilcorp'),
    'job': fields.String(default='fsociety'),
    'mountpoint': fields.String(default='/'),
    'severity': fields.String(default='severe')
})

alert = api_v1.model('alert', {
    'status': fields.String(default='Resolved'),
    'startsAt': fields.String(default='2017-05-09T07:01:37.803Z'),
    'endsAt': fields.String(default='2017-05-09T07:01:37.803Z'),
    'generatorURL': fields.String(default='my.prometheusserver.url'),
    'labels': fields.Nested(labels),
    'annotations': fields.Nested(annotations)
})

message = api_v1.model('message', {
    'receiver': fields.String(default='test_webhook'),
    'status': fields.String(default='Resolved'),
    'alerts': fields.List(fields.Nested(alert)),
    'externalURL': fields.String(default='my.prometheusalertmanager.url'),
    'version': fields.String(default='4')
})



import argparse
import configparser
import logging.config

import os

import sys
import yaml

from prom2teams import root
from prom2teams.app.exceptions import MissingConnectorConfigKeyException


def _config_command_line():
    parser = argparse.ArgumentParser(description='Receives alert notifications '
                                                 'from Prometheus Alertmanager '
                                                 '(https://github.com/prometheus/alertmanager) '
                                                 'and sends it to Microsoft Teams using configured connectors ')

    parser.add_argument('-c', '--configpath', help='config INI file path', required=False)
    parser.add_argument('-g', '--groupalertsby', help='group alerts with same attribute into one alarm', required=False)
    parser.add_argument('-l', '--logfilepath', help='log file path', required=False)
    parser.add_argument('-v', '--loglevel', help='log level', required=False)
    parser.add_argument('-t', '--templatepath', help='Jinja2 template file path', required=False)
    parser.add_argument('-s', '--labelsexcluded', help='prometheus custom labels to be ignored', required=False)
    parser.add_argument('-a', '--annotationsexcluded', help='prometheus custom annotations to be ignored', required=False)
    parser.add_argument('-m', '--enablemetrics', action='store_true', help='enable Prom2teams Prometheus metrics', required=False)
    return parser.parse_args()


def _update_application_configuration(application, configuration):
    if 'Microsoft Teams' in configuration:
        application.config['MICROSOFT_TEAMS'] = configuration['Microsoft Teams']
    if 'Template' in configuration and 'Path' in configuration['Template']:
        application.config['TEMPLATE_PATH'] = configuration['Template']['Path']
    if 'Log' in configuration and 'Level' in configuration['Log']:
        application.config['LOG_LEVEL'] = configuration['Log']['Level']
    if 'Log' in configuration and 'Path' in configuration['Log']:
        application.config['LOG_FILE_PATH'] = configuration['Log']['Path']
    if 'Group Alerts' in configuration:
        application.config['GROUP_ALERTS_BY'] = configuration['Group Alerts']['Field']
    if 'HTTP Server' in configuration:
        if 'Host' in configuration['HTTP Server']:
            _host = configuration['HTTP Server']['Host']
            application.config['HOST'] = _host
        if 'Port' in configuration['HTTP Server']:
            _port = configuration['HTTP Server']['Port']
            application.config['PORT'] = _port
    if 'Labels' in configuration:
        application.config['LABELS_EXCLUDED'] = tuple(configuration['Labels']['Excluded'].replace(' ', '').split(','))
    if 'Annotations' in configuration:
        application.config['ANNOTATIONS_EXCLUDED'] = tuple(configuration['Annotations']['Excluded'].replace(' ', '').split(','))


def _config_provided(filepath):
    config = configparser.ConfigParser()
    try:
        with open(filepath) as f_prov:
            config.read_file(f_prov)

        if not config.options('Microsoft Teams'):
            raise MissingConnectorConfigKeyException('missing connector key in provided config')

    except configparser.NoSectionError:
        raise MissingConnectorConfigKeyException('missing required Microsoft Teams / '
                                                 'connector key in provided config')
    return config


def setup_logging(application):
    with open(os.path.join(root, 'config/logging.yml'), 'rt') as f:
        config = yaml.safe_load(f.read())

        for logger in config['loggers']:
            config['loggers'][logger]['level'] = application.config['LOG_LEVEL']
        config['root']['level'] = application.config['LOG_LEVEL']
        config['loggers']['prom2teams_app']['level'] = 'INFO'

        environment = os.getenv('APP_ENVIRONMENT', 'None')
        if environment == 'pro' or environment == 'pre':
            config['handlers']['file']['filename'] = application.config['LOG_FILE_PATH']
            for logger in config['loggers']:
                config['loggers'][logger]['handlers'] = ['file']
            config['root']['handlers'] = ['file']
        else:
            del config['handlers']['file']

        logging.config.dictConfig(config)


def config_app(application):
    try:
        # Load the default configuration
        application.config.from_object('prom2teams.config.settings')

        # Load the configuration from the instance folder
        instance = os.path.join(os.path.join(root, os.pardir), 'instance')
        config = os.path.join(instance, 'config.py')
        if os.path.isdir(instance) and os.path.exists(config):
            application.config.from_pyfile(config)

        # Load the file specified by the APP_CONFIG_FILE environment variable
        # Variables defined here will override those in the default configuration
        if 'APP_CONFIG_FILE' in os.environ:
            application.config['APP_CONFIG_FILE'] = os.environ.get('APP_CONFIG_FILE')
            config_provided = _config_provided(os.getenv('APP_CONFIG_FILE'))
            _update_application_configuration(application, config_provided)

        # Parse and load command line properties
        # Variables defined here will override previous configuration
        command_line_args = _config_command_line()
        if command_line_args.configpath:
            application.config['APP_CONFIG_FILE'] = command_line_args.configpath
            config_provided = _config_provided(command_line_args.configpath)
            _update_application_configuration(application, config_provided)
        if command_line_args.loglevel:
            application.config['LOG_LEVEL'] = command_line_args.loglevel
        if command_line_args.logfilepath:
            application.config['LOG_FILE_PATH'] = command_line_args.logfilepath
        if command_line_args.templatepath:
            application.config['TEMPLATE_PATH'] = command_line_args.templatepath
        if command_line_args.groupalertsby:
            application.config['GROUP_ALERTS_BY'] = command_line_args.groupalertsby
        if command_line_args.enablemetrics or os.environ.get('PROM2TEAMS_PROMETHEUS_METRICS', False):
            from prometheus_flask_exporter import PrometheusMetrics
            metrics = PrometheusMetrics(application)

        if 'MICROSOFT_TEAMS' not in application.config:
            raise MissingConnectorConfigKeyException('missing connector key in config')

    except MissingConnectorConfigKeyException:
        sys.exit('No Microsoft Teams connector available')



import requests

from .exceptions import MicrosoftTeamsRequestException

session = requests.Session()
session.headers.update({'Content-Type': 'application/json'})


def post(teams_webhook_url, message):
    response = session.post(teams_webhook_url, data=message)
    if not response.ok or response.text is not '1':
        exception_msg = 'Error performing request to: {}.\n' \
                        ' Returned status code: {}.\n' \
                        ' Returned data: {}\n' \
                        ' Sent message: {}\n'
        raise MicrosoftTeamsRequestException(exception_msg.format(teams_webhook_url,
                                                                  str(response.status_code),
                                                                  str(response.text),
                                                                  str(message)),
                                             code=response.status_code)



from prom2teams.teams.teams_alarm_schema import TeamsAlarm, TeamsAlarmSchema
from collections import defaultdict


def map_prom_alerts_to_teams_alarms(alerts):
    alerts = group_alerts(alerts, 'status')
    teams_alarms = []
    schema = TeamsAlarmSchema()
    for same_status_alerts in alerts:
        for alert in alerts[same_status_alerts]:
            alarm = TeamsAlarm(alert.name, alert.status.lower(), alert.severity,
                               alert.summary, alert.instance, alert.description, alert.extra_labels, alert.extra_annotations)
            json_alarm = schema.dump(alarm)
            teams_alarms.append(json_alarm)
    return teams_alarms


def map_and_group(alerts, group_alerts_by):
    alerts = group_alerts(alerts, 'status')
    teams_alarms = []
    schema = TeamsAlarmSchema()
    for same_status_alerts in alerts:
        grouped_alerts = group_alerts(alerts[same_status_alerts], group_alerts_by)
        for alert in grouped_alerts:
            features = group_features(grouped_alerts[alert])
            name, description, instance, severity, status, summary = (teams_visualization(features["name"]),
                                                                      teams_visualization(features["description"]),
                                                                      teams_visualization(features["instance"]),
                                                                      teams_visualization(features["severity"]),
                                                                      teams_visualization(features["status"]),
                                                                      teams_visualization(features["summary"]))
            extra_labels = dict()
            extra_annotations = dict()
            for element in grouped_alerts[alert]:
                if hasattr(element, 'extra_labels'):
                    extra_labels = {**extra_labels, **element.extra_labels}
                if hasattr(element, 'extra_annotations'):
                    extra_annotations = {**extra_annotations, **element.extra_annotations}

            alarm = TeamsAlarm(name, status.lower(), severity, summary,
                               instance, description, extra_labels, extra_annotations)
            json_alarm = schema.dump(alarm)
            teams_alarms.append(json_alarm)
    return teams_alarms


def teams_visualization(feature):
    feature.sort()
    # Teams won't print just one new line
    return ',\n\n\n'.join(feature) if feature else 'unknown'


def group_alerts(alerts, group_alerts_by):
    groups = defaultdict(list)
    for alert in alerts:
        groups[alert.__dict__[group_alerts_by]].append(alert)
    return dict(groups)


def group_features(alerts):
    grouped_features = {feature: list(set([individual_alert.__dict__[feature] for individual_alert in alerts]))
                        for feature in ["name", "description", "instance", "severity", "status", "summary"]}
    return grouped_features



import warnings

from flask import request, current_app as app
from flask_restplus import Resource

from prom2teams.app.sender import AlarmSender
from prom2teams.prometheus.message_schema import MessageSchema
from .model import *

ns = api_v1.namespace(name='', description='Version 1 connections')


@ns.route('/')
@api_v1.doc(responses={201: 'OK'})
class AlertReceiver(Resource):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.schema = MessageSchema()
        if 'TEMPLATE_PATH' in app.config:
            self.sender = AlarmSender(app.config['TEMPLATE_PATH'])
        else:
            self.sender = AlarmSender()

    @api_v1.expect(message)
    def post(self):
        _show_deprecated_warning("Call to deprecated function. It will be removed in future versions. "
                                 "Please view the README file.")
        alerts = self.schema.load(request.get_json())
        self.sender.send_alarms(alerts, app.config['MICROSOFT_TEAMS']['Connector'])
        return 'OK', 201


def _show_deprecated_warning(msg):
    warnings.warn(message=msg, category=PendingDeprecationWarning)



import logging
import os

from jinja2 import Environment, FileSystemLoader

from prom2teams import root
from prom2teams.app.exceptions import MissingTemplatePathException


log = logging.getLogger('prom2teams')


class _Singleton(type):

    def __init__(cls, name, bases, attrs, **kwargs):
        super().__init__(name, bases, attrs)
        cls._instance = None

    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__call__(*args, **kwargs)
        return cls._instance


class TemplateComposer(metaclass=_Singleton):

    DEFAULT_TEMPLATE_PATH = os.path.abspath(os.path.join(root, 'resources/templates/teams.j2'))

    def __init__(self, template_path=DEFAULT_TEMPLATE_PATH):
        log.info(template_path)
        if not os.path.isfile(template_path):
            raise MissingTemplatePathException('Template {} not exists'.format(template_path))

        template_dir = os.path.dirname(template_path)
        template_name = os.path.basename(template_path)
        loader = FileSystemLoader(template_dir)
        environment = Environment(loader=loader, trim_blocks=True)
        self.template = environment.get_template(template_name)

    def compose_all(self, alarms_json):
        rendered_templates = [self.template.render(status=json_alarm['status'], msg_text=json_alarm)
                              for json_alarm in alarms_json]
        return rendered_templates



from flask import request, current_app as app
from flask_restplus import Resource

from prom2teams.app.sender import AlarmSender
from prom2teams.prometheus.message_schema import MessageSchema
from .model import *
from marshmallow import EXCLUDE
ns = api_v2.namespace(name='', description='Version 2 connections')


@ns.route('/<string:connector>')
@api_v2.doc(params={'connector': 'Name of connector to use'},
            responses={201: 'OK'})
class AlertReceiver(Resource):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.schema = MessageSchema(exclude_fields=app.config['LABELS_EXCLUDED'], exclude_annotations=app.config['ANNOTATIONS_EXCLUDED'], unknown=EXCLUDE)
        if app.config['TEMPLATE_PATH']:
            self.sender = AlarmSender(app.config['TEMPLATE_PATH'], app.config['GROUP_ALERTS_BY'])
        else:
            self.sender = AlarmSender(group_alerts_by=app.config['GROUP_ALERTS_BY'])

    @api_v2.expect(message)
    def post(self, connector):
        alerts = self.schema.load(request.get_json())
        self.sender.send_alarms(alerts, app.config['MICROSOFT_TEAMS'][connector])
        return 'OK', 201



import logging

from prom2teams.teams.alarm_mapper import map_and_group, map_prom_alerts_to_teams_alarms
from prom2teams.teams.composer import TemplateComposer
from .teams_client import post

log = logging.getLogger('prom2teams')


class AlarmSender:

    def __init__(self, template_path=None, group_alerts_by=False):
        self.group_alerts_by = group_alerts_by
        if template_path:
            self.json_composer = TemplateComposer(template_path)
        else:
            self.json_composer = TemplateComposer()

    def _create_alarms(self, alerts):
        if self.group_alerts_by:
            alarms = map_and_group(alerts, self.group_alerts_by)
        else:
            alarms = map_prom_alerts_to_teams_alarms(alerts)
        return self.json_composer.compose_all(alarms)

    def send_alarms(self, alerts, teams_webhook_url):
        sending_alarms = self._create_alarms(alerts)
        for team_alarm in sending_alarms:
            log.debug('The message that will be sent is: %s', str(team_alarm))
            post(teams_webhook_url, team_alarm)



from marshmallow import Schema, fields, INCLUDE


class TeamsAlarmSchema(Schema):
    class Meta:
        unknown = INCLUDE
    status = fields.Str()
    severity = fields.Str()
    summary = fields.Str()
    instance = fields.Str()
    description = fields.Str()
    name = fields.Str()
    extra_labels = fields.Dict(
        keys=fields.Str(),
        values=fields.Str(),
        required=False,
        data_key='extra_labels'
    )
    extra_annotations = fields.Dict(
        keys=fields.Str(),
        values=fields.Str(),
        required=False,
        data_key='extra_annotations'
    )


class TeamsAlarm:
    def __init__(self, name, status, severity, summary, instance, description, extra_labels, extra_annotations):
        self.name = name
        self.status = status
        self.severity = severity
        self.summary = summary
        self.instance = instance
        self.description = description
        self.extra_labels = extra_labels
        self.extra_annotations = extra_annotations



APP_NAME = 'prom2teams'
HOST = 'localhost'
PORT = 8089
# Flask settings
DEBUG = False

# Flask-Restplus settings
SWAGGER_UI_DOC_EXPANSION = 'list'
SWAGGER_UI_JSONEDITOR = True
VALIDATE = True
MASK_SWAGGER = False
ERROR_404_HELP = False

# Logging extra settings
LOG_FILE_PATH = '/var/log/' + APP_NAME + '/' + APP_NAME + '.log'
LOG_LEVEL = 'DEBUG'

# Group alerts setting
GROUP_ALERTS_BY = False

# Api
API_V1_URL_PREFIX = ''
API_V2_URL_PREFIX = '/v2'

# Template
TEMPLATE_PATH = None

# Labels
LABELS_EXCLUDED = ()

# Annotations
ANNOTATIONS_EXCLUDED = ()



import sys
import os


try:
    from prom2teams.app.api import app as application
except ImportError:
    sys.path.append(os.path.abspath('./'))
    from prom2teams.app.api import app as application


if __name__ == "__main__":
    host = application.config['HOST']
    port = int(application.config['PORT'])
    application.run(host=host, port=port)



from setuptools import setup, find_packages


with open('requirements.txt') as req:
    requirements = req.read().splitlines()


with open('README.md', encoding='utf-8') as f:
    readme = f.read()


setup(name='prom2teams',
      version='2.5.7',
      description='Project that redirects Prometheus Alert Manager '
      'notifications to Microsoft Teams',
      long_description=readme,
      long_description_content_type='text/markdown',
      install_requires=requirements,
      setup_requires=[
        'flake8'
      ],
      scripts=[
          'bin/prom2teams',
          'bin/prom2teams_uwsgi'
      ],
      package_data={
        '': ['*.ini', '*.j2', '*.ico'],
      },
      include_package_data=True,
      data_files=[
          ('/usr/local/etc/prom2teams', ['bin/wsgi.py'])
      ],
      url='https://github.com/idealista/prom2teams',
      author='Idealista, S.A.U',
      author_email='labs@idealista.com',
      license='Apache license 2.0',
      packages=find_packages(exclude=('tests', 'docs')),
      keywords='microsoft teams prometheus alert',
      classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Utilities',
        'Topic :: Communications :: Chat',
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
      ],
      zip_safe=False)



import logging.config
import os

from flask import Flask, Blueprint, send_from_directory

from prom2teams.app.configuration import config_app, setup_logging
from .versions.v1 import api_v1
from .versions.v1.namespace import ns as ns_v1
from .versions.v2 import api_v2
from .versions.v2.namespace import ns as ns_v2

log = logging.getLogger('prom2teams_app')

app = Flask(__name__)



@app.route('/favicon.ico')
def favicon():
    return send_from_directory(os.path.join(app.root_path, 'static'),
                               'favicon.ico', mimetype='image/vnd.microsoft.icon')

def error_handler(e):
    msg = 'An unhandled exception occurred. {}'.format(e)
    log.exception(msg)
    return str(e), e.code



def register_api(application, api, namespace, blueprint):
    api.init_app(blueprint)
    api.add_namespace(namespace)
    application.register_blueprint(blueprint)


def init_app(application):
    config_app(application)
    setup_logging(application)

    blueprint_v1 = Blueprint('api_v1', __name__, url_prefix=application.config['API_V1_URL_PREFIX'])
    blueprint_v2 = Blueprint('api_v2', __name__, url_prefix=application.config['API_V2_URL_PREFIX'])
    register_api(application, api_v1, ns_v1, blueprint_v1)
    register_api(application, api_v2, ns_v2, blueprint_v2)
    application.register_error_handler(500, error_handler)


init_app(app)
log.info('{} started on {}:{}'.format(app.config['APP_NAME'], app.config['HOST'], app.config['PORT']))



class MicrosoftTeamsRequestException(Exception):
    def __init__(self, msg, code=None):
        self.code = code


class MissingConnectorConfigKeyException(Exception):
    pass


class MissingTemplatePathException(Exception):
    pass
